# Test class definition
class Account
  getter name : String
  getter balance : Float64

  def initialize(@name, @balance)
  end

  def deposit(amount)
    @balance += amount
  end

  def withdraw(amount)
    @balance -= amount
  end
end

# Register and bind Crystal methods for Lua
# This is a great opportunity for macros using type introspection
class Account
  def self.register(lua)
    methods = [
      LibLua::Reg.new(name: "name", func: ->(lua : LibLua::State) {
        acct = LibLua.checkudata(lua, 1, "Account").as(Account)
        LibLua.pushstring(lua, acct.name)
        1
      }),
      LibLua::Reg.new(name: "balance", func: ->(lua : LibLua::State) {
        acct = LibLua.checkudata(lua, 1, "Account").as(Account)
        LibLua.pushnumber(lua, acct.balance)
        1
      }),
      LibLua::Reg.new(name: "deposit", func: ->(lua : LibLua::State) {
        acct = LibLua.checkudata(lua, 1, "Account").as(Account)
        amount = LibLua.checknumber(lua, 2)
        new_balance = acct.deposit(amount)
        LibLua.pushnumber(lua, new_balance)
        1
      }),
      LibLua::Reg.new(name: "withdraw", func: ->(lua : LibLua::State) {
        acct = LibLua.checkudata(lua, 1, "Account").as(Account)
        amount = LibLua.checknumber(lua, 2)
        new_balance = acct.withdraw(amount)
        LibLua.pushnumber(lua, new_balance)
        1
      }),
      LibLua::Reg.new,
    ]

    LibLua.newmetatable(lua, "Account")
    LibLua.setfuncs(lua, methods, 0)
    LibLua.pushvalue(lua, -1)
    LibLua.setfield(lua, -1, "__index")
    LibLua.setglobal(lua, "Account")
  end

  def to_unsafe
    self.as(Void*)
  end
end

my_account = Account.new(name: "My Account", balance: 100.0)

lua = LibLua.newstate
LibLua.openlibs(lua)

Account.register(lua)

LibLua.pushlightuserdata(lua, my_account)
LibLua.getfield(lua, LibLua::REGISTRY_INDEX, "Account")
LibLua.setmetatable(lua, -2)
LibLua.setglobal(lua, "my_account")

result = dostring(lua, File.new("account.lua").gets_to_end)
if result != 0
  puts "Failed to execute file"
end

LibLua.close(lua)

@[Link("lua")]
lib LibLua
  MULTRET = -1

  REGISTRY_INDEX = -1001000

  type State = Void*
  alias Func = State -> Int32
  alias Integer = LibC::Long
  alias Number = Float64

  struct Reg
    name : UInt8*
    func : Func
  end

  fun newstate = luaL_newstate : State
  fun openlibs = luaL_openlibs(l : State) : Void

  fun loadstring = luaL_loadstring(l : State, s : UInt8*) : Int32
  fun callk = lua_callk(l : State, nargs : Int32, nret : Int32, ctx : Int32, k : Func) : Int32
  fun pcallk = lua_pcallk(l : State, nargs : Int32, nret : Int32, errfunc : Int32, ctx : Int32, k : Func) : Int32

  fun getglobal = lua_getglobal(l : State, name : UInt8*) : Void
  fun setglobal = lua_setglobal(l : State, name : UInt8*) : Void
  fun pushinteger = lua_pushinteger(l : State, n : Integer) : Void
  fun pushnumber = lua_pushnumber(l : State, n : Number) : Void
  fun pushcclosure = lua_pushcclosure(l : State, f : Func, nup : Int32) : Void
  fun pushlightuserdata = lua_pushlightuserdata(l : State, data : Void*) : Void
  fun pushstring = lua_pushstring(l : State, s : UInt8*) : UInt8*
  fun pushvalue = lua_pushvalue(l : State, idx : Int32) : Void
  fun checknumber = luaL_checknumber(l : State, arg : Int32) : Number
  fun checkudata = luaL_checkudata(l : State, arg : Int32, tname : UInt8*) : Void*
  fun tointegerx = lua_tointegerx(l : State, idx : Int32, isnum : Int32*) : Integer

  fun newmetatable = luaL_newmetatable(l : State, tname : UInt8*) : Int32
  fun setfuncs = luaL_setfuncs(l : State, methods : Reg*, nup : Int32) : Void
  fun setfield = lua_setfield(l : State, idx : Int32, key : UInt8*) : Void
  fun getfield = lua_getfield(l : State, idx : Int32, tname : UInt8*) : Void
  fun setmetatable = lua_setmetatable(l : State, idx : Int32) : Void

  fun close = lua_close(l : State) : Void
end

def dostring(lua, s)
  if (result = LibLua.loadstring(lua, s)) == 0
    result = LibLua.pcallk(lua, 0, LibLua::MULTRET, 0, 0, nil)
  end
  result
end
